"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("underscore");
const model = require("./model");
const utils_1 = require("@nodeswork/utils");
const _1 = require("./");
const dotty = require('dotty');
exports.READONLY = 'READONLY';
exports.AUTOGEN = 'AUTOGEN';
class KoaMiddlewares extends model.DocumentModel {
    static createMiddleware(options) {
        const self = this.cast();
        _.defaults(options, DEFAULT_COMMON_OPTIONS);
        async function create(ctx, next) {
            const opts = _.extend({}, options, ctx.overrides && ctx.overrides.options);
            let rModel = self;
            const omits = _.union(['_id'], opts.omits, self.schema.api.AUTOGEN);
            let doc = _.omit(ctx.request.body, omits);
            doc = _.extend(doc, ctx.overrides && ctx.overrides.doc);
            ctx[opts.target] = doc;
            if (opts.triggerNext) {
                await next();
            }
            doc = ctx[opts.target];
            let object = (await rModel.create(doc));
            if (opts.project || opts.level) {
                object = await self.findById(object._id, opts.project, {
                    level: opts.level,
                });
            }
            ctx[opts.target] = object;
            if (opts.populate) {
                await rModel.populate(object, opts.populate);
            }
            if (!opts.noBody) {
                ctx.body = await opts.transform(object, ctx);
            }
        }
        Object.defineProperty(create, 'name', {
            value: `${self.modelName}#createMiddleware`,
            writable: false,
        });
        return create;
    }
    static getMiddleware(options) {
        const self = this.cast();
        _.defaults(options, DEFAULT_COMMON_OPTIONS);
        async function get(ctx, next) {
            const opts = _.extend({}, options, ctx.overrides && ctx.overrides.options);
            const query = (ctx.overrides && ctx.overrides.query) || {};
            if (opts.field !== '*') {
                if (opts.field.indexOf('.') >= 0) {
                    query._id = dotty.get(ctx.request, opts.field);
                }
                else {
                    query._id = ctx.params[opts.field];
                }
                if (query._id == null) {
                    throw new utils_1.NodesworkError('invalid value', {
                        responseCode: 422,
                        path: opts.field,
                    });
                }
            }
            if (Object.keys(query).length === 0) {
                throw new utils_1.NodesworkError('no query parameters', {
                    responseCode: 422,
                    path: opts.field,
                });
            }
            const queryOption = {};
            if (opts.level) {
                queryOption.level = opts.level;
            }
            let queryPromise = self.findOne(query, opts.project, queryOption);
            if (opts.populate) {
                queryPromise = queryPromise.populate(opts.populate);
            }
            const object = await queryPromise;
            ctx[opts.target] = object;
            if (!opts.nullable && object == null) {
                throw utils_1.NodesworkError.notFound();
            }
            if (opts.triggerNext) {
                await next();
            }
            if (!opts.noBody) {
                ctx.body = await opts.transform(ctx[opts.target], ctx);
            }
        }
        Object.defineProperty(get, 'name', {
            value: `${self.modelName}#getMiddleware`,
            writable: false,
        });
        return get;
    }
    static findMiddleware(options) {
        const self = this.cast();
        _.defaults(options, DEFAULT_COMMON_OPTIONS);
        if (options.pagination) {
            _.defaults(options.pagination, DEFAULT_FIND_PAGINATION_OPTIONS);
        }
        const defaultPagination = {
            page: 0,
            size: options.pagination ? options.pagination.size : 0,
        };
        async function find(ctx, next) {
            const opts = _.extend({}, options, ctx.overrides && ctx.overrides.options);
            const query = (ctx.overrides && ctx.overrides.query) || {};
            const queryOption = {};
            let pagination = null;
            if (opts.pagination) {
                pagination = ctx.request.query;
                _.defaults(pagination, defaultPagination);
                queryOption.skip = pagination.page * pagination.size;
                queryOption.limit = pagination.size;
            }
            if (opts.sort) {
                queryOption.sort = opts.sort;
            }
            if (ctx.overrides && ctx.overrides.sort) {
                queryOption.sort = ctx.overrides.sort;
            }
            if (opts.level) {
                queryOption.level = opts.level;
            }
            let queryPromise = self.find(query, opts.project, queryOption);
            if (opts.populate) {
                queryPromise = queryPromise.populate(opts.populate);
            }
            const object = await queryPromise;
            ctx[opts.target] = object;
            if (opts.triggerNext) {
                await next();
            }
            if (pagination) {
                const totalPage = Math.floor(((await self.find(query).count()) - 1) / pagination.size + 1);
                ctx.response.set('total_page', totalPage.toString());
            }
            if (!opts.noBody) {
                const body = ctx[opts.target];
                for (let i = 0; i < body.length; i++) {
                    body[i] = await opts.transform(body[i], ctx);
                }
                ctx.body = body;
            }
        }
        Object.defineProperty(find, 'name', {
            value: `${self.modelName}#findMiddleware`,
            writable: false,
        });
        return find;
    }
    static updateMiddleware(options) {
        const self = this.cast();
        _.defaults(options, DEFAULT_COMMON_OPTIONS);
        async function update(ctx, next) {
            const opts = _.extend({}, options, ctx.overrides && ctx.overrides.options);
            const query = (ctx.overrides && ctx.overrides.query) || {};
            if (opts.field !== '*') {
                if (opts.field.indexOf('.') >= 0) {
                    query._id = dotty.get(ctx.request, opts.field);
                }
                else {
                    query._id = ctx.params[opts.field];
                }
                if (query._id == null) {
                    throw new utils_1.NodesworkError('invalid value', {
                        responseCode: 422,
                        path: opts.field,
                    });
                }
            }
            if (Object.keys(query).length === 0) {
                throw new utils_1.NodesworkError('no query parameters', {
                    responseCode: 422,
                    path: opts.field,
                });
            }
            const queryOption = {
                new: true,
                fields: opts.project,
                level: opts.level,
                runValidators: true,
            };
            const omits = _.union(['_id'], opts.omits, self.schema.api.READONLY, self.schema.api.AUTOGEN);
            let doc = _.omit(ctx.request.body, omits);
            doc = _.extend(doc, ctx.overrides && ctx.overrides.doc);
            const upDoc = {
                $set: doc,
            };
            let updatePromise = self.findOneAndUpdate(query, upDoc, queryOption);
            if (opts.populate) {
                updatePromise = updatePromise.populate(opts.populate);
            }
            const object = await updatePromise;
            ctx[opts.target] = object;
            if (opts.triggerNext) {
                await next();
            }
            if (!opts.noBody) {
                ctx.body = await opts.transform(ctx[opts.target], ctx);
            }
        }
        Object.defineProperty(update, 'name', {
            value: `${self.modelName}#updateMiddleware`,
            writable: false,
        });
        return update;
    }
    static deleteMiddleware(options) {
        const self = this.cast();
        _.defaults(options, DEFAULT_COMMON_OPTIONS);
        async function del(ctx, next) {
            const opts = _.extend({}, options, ctx.overrides && ctx.overrides.options);
            const query = (ctx.overrides && ctx.overrides.query) || {};
            if (opts.field.indexOf('.') >= 0) {
                query._id = dotty.get(ctx.request, opts.field);
            }
            else {
                query._id = ctx.params[opts.field];
            }
            const queryOption = {};
            const queryPromise = self.findOne(query, undefined, queryOption);
            let object = await queryPromise;
            ctx[opts.target] = object;
            object = ctx[opts.target];
            if (!opts.nullable && object == null) {
                throw new utils_1.NodesworkError('not found', {
                    responseCode: 404,
                });
            }
            if (object) {
                await object.remove();
            }
            if (opts.triggerNext) {
                await next();
            }
            ctx.status = 204;
        }
        Object.defineProperty(del, 'name', {
            value: `${self.modelName}#deleteMiddleware`,
            writable: false,
        });
        return del;
    }
}
exports.KoaMiddlewares = KoaMiddlewares;
const DEFAULT_COMMON_OPTIONS = {
    target: 'object',
    transform: _.identity,
};
const DEFAULT_FIND_PAGINATION_OPTIONS = {
    size: 20,
    sizeChoices: [20, 50, 100, 200],
};
KoaMiddlewares.Plugin({
    fn: apiLevel,
});
function apiLevel(schema, options) {
    for (let s = schema; s; s = s.parentSchema) {
        if (s.api == null) {
            s.api = {
                READONLY: [],
                AUTOGEN: [],
            };
        }
    }
    schema.eachPath((pathname, schemaType) => {
        if ([exports.READONLY, exports.AUTOGEN].indexOf(schemaType.options.api) < 0) {
            return;
        }
        for (let s = schema; s != null; s = s.parentSchema) {
            s.api[schemaType.options.api] = _.union(s.api[schemaType.options.api], [
                pathname,
            ]);
        }
    });
}
function Autogen(schema = {}) {
    return _1.Field(_.extend({}, schema, {
        api: exports.AUTOGEN,
    }));
}
exports.Autogen = Autogen;
function Readonly(schema = {}) {
    return _1.Field(_.extend({}, schema, {
        api: exports.READONLY,
    }));
}
exports.Readonly = Readonly;

//# sourceMappingURL=koa.js.map
